"""
This script verifies whether UXsim outputs reasonable solutions for multilane scenarios.
Some of the tests are probablistic.
"""

import pytest
import random
from uxsim import *
import numpy as np

def equal_tolerance(val, check, rel_tol=0.1, abs_tol=0.0):
    if check == 0 and abs_tol == 0:
        abs_tol = 0.1
    return abs(val - check) <= abs(check*rel_tol) + abs_tol

"""
default FD:
    u = 20
    kappa = 0.2
    tau = 1
    w = 5
    k^* = 0.04
    q^* = 0.8
"""

def q2k_cong(q, link):
    return -(q-link.kappa*link.w)/link.w



@pytest.mark.flaky(reruns=5)
def test_straight_1link_2lane_low_demand():
    W = World(
        name="",
        deltan=5,
        tmax=1200,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("dest", 2, 1)
    link = W.addLink("link", "orig", "dest", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    t = 600
    flow = 1
    W.adddemand("orig", "dest", 0, t, flow)

    W.exec_simulation()
    W.analyzer.print_simple_stats()
    W.analyzer.compute_edie_state()
    
    assert equal_tolerance(link.capacity, 0.8*2)
    assert equal_tolerance(link.u, 20)
    assert equal_tolerance(link.k_star, 0.04*2)
    assert equal_tolerance(link.q_mat[2:4,3:5].mean(), 1)
    assert equal_tolerance(link.v_mat[2:4,3:5].mean(), link.u)
    assert equal_tolerance(link.k_mat[2:4,3:5].mean(), 1/link.u)
    assert equal_tolerance(link.q_mat[6:,3:5].mean(), 0)
    assert equal_tolerance(link.v_mat[6:,3:5].mean(), link.u)
    assert equal_tolerance(link.k_mat[6:,3:5].mean(), 0)
    
@pytest.mark.flaky(reruns=5)
def test_straight_1link_2lane_high_demand():
    W = World(
        name="",
        deltan=5,
        tmax=1200,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("dest", 2, 1)
    link = W.addLink("link", "orig", "dest", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    t = 1200
    flow = 2
    W.adddemand("orig", "dest", 0, t, flow)

    W.exec_simulation()
    W.analyzer.print_simple_stats()
    W.analyzer.compute_edie_state()
    
    assert equal_tolerance(link.capacity, 0.8*2)
    assert equal_tolerance(link.u, 20)
    assert equal_tolerance(link.k_star, 0.04*2)
    assert equal_tolerance(link.q_mat[2:6,3:5].mean(), link.capacity)
    assert equal_tolerance(link.v_mat[2:6,3:5].mean(), link.u)
    assert equal_tolerance(link.k_mat[2:6,3:5].mean(), link.k_star)
    

@pytest.mark.flaky(reruns=5)
def test_straight_2link_2lane_low_demand():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    t = 600
    flow = 1
    W.adddemand("orig", "dest", 0, t, flow)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
        
    assert equal_tolerance(link1.q_mat[2:4,3:5].mean(), 1)
    assert equal_tolerance(link1.v_mat[2:4,3:5].mean(), link1.u)
    assert equal_tolerance(link1.k_mat[2:4,3:5].mean(), 1/link1.u)
    assert equal_tolerance(link1.q_mat[7:,3:5].mean(), 0)
    assert equal_tolerance(link1.v_mat[7:,3:5].mean(), link1.u)
    assert equal_tolerance(link1.k_mat[7:,3:5].mean(), 0)
    assert equal_tolerance(link2.q_mat[2:4,3:5].mean(), 1)
    assert equal_tolerance(link2.v_mat[2:4,3:5].mean(), link2.u)
    assert equal_tolerance(link2.k_mat[2:4,3:5].mean(), 1/link2.u)
    assert equal_tolerance(link2.q_mat[7:,3:5].mean(), 0)
    assert equal_tolerance(link2.v_mat[7:,3:5].mean(), link2.u)
    assert equal_tolerance(link2.k_mat[7:,3:5].mean(), 0)
    

@pytest.mark.flaky(reruns=5)
def test_straight_2link_2lane_high_demand():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    t = 400
    flow = 2
    W.adddemand("orig", "dest", 0, t, flow)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
        
    assert equal_tolerance(link1.q_mat[2:4,3:5].mean(), 0.8*2)
    assert equal_tolerance(link1.v_mat[2:4,3:5].mean(), 20)
    assert equal_tolerance(link1.k_mat[2:4,3:5].mean(), 0.04*2)
    assert equal_tolerance(link1.q_mat[7:,3:5].mean(), 0)
    assert equal_tolerance(link1.v_mat[7:,3:5].mean(), link1.u)
    assert equal_tolerance(link1.k_mat[7:,3:5].mean(), 0)
    assert equal_tolerance(link2.q_mat[2:4,3:5].mean(), 0.8*2)
    assert equal_tolerance(link2.v_mat[2:4,3:5].mean(), 20)
    assert equal_tolerance(link2.k_mat[2:4,3:5].mean(), 0.04*2)
    assert equal_tolerance(link2.q_mat[7:,3:5].mean(), 0)
    assert equal_tolerance(link2.v_mat[7:,3:5].mean(), link2.u)
    assert equal_tolerance(link2.k_mat[7:,3:5].mean(), 0)
    


@pytest.mark.flaky(reruns=5)
def test_straight_2link_2lane_to_1lane_congestion():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
    t = 1000
    flow = 2
    W.adddemand("orig", "dest", 0, t, flow)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
        
    assert equal_tolerance(link1.q_mat[2:4,3:5].mean(), 0.8*2)  #before congestion
    assert equal_tolerance(link1.v_mat[2:4,3:5].mean(), 20)
    assert equal_tolerance(link1.k_mat[2:4,3:5].mean(), 0.04*2)
    assert equal_tolerance(link1.q_mat[7:,3:5].mean(), 0.8)  #after congestion
    assert equal_tolerance(link1.v_mat[7:,3:5].mean(), 3.33)
    assert equal_tolerance(link1.k_mat[7:,3:5].mean(), 0.24)
    assert equal_tolerance(link2.q_mat[2:,3:5].mean(), 0.8)  #downstream
    assert equal_tolerance(link2.v_mat[2:,3:5].mean(), 20)
    assert equal_tolerance(link2.k_mat[2:,3:5].mean(), 0.04)
    

@pytest.mark.flaky(reruns=5)
def test_straight_2link_2lane_to_1lane_congestion_deltan1():
    W = World(
        name="",
        deltan=1,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
    t = 1000
    flow = 2
    W.adddemand("orig", "dest", 0, t, flow)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
        
    assert equal_tolerance(link1.q_mat[2:4,3:5].mean(), 0.8*2)  #before congestion
    assert equal_tolerance(link1.v_mat[2:4,3:5].mean(), 20)
    assert equal_tolerance(link1.k_mat[2:4,3:5].mean(), 0.04*2)
    assert equal_tolerance(link1.q_mat[7:,3:5].mean(), 0.8)  #after congestion
    assert equal_tolerance(link1.v_mat[7:,3:5].mean(), 3.33)
    assert equal_tolerance(link1.k_mat[7:,3:5].mean(), 0.24)
    assert equal_tolerance(link2.q_mat[2:,3:5].mean(), 0.8)  #downstream
    assert equal_tolerance(link2.v_mat[2:,3:5].mean(), 20)
    assert equal_tolerance(link2.k_mat[2:,3:5].mean(), 0.04)
    
    

@pytest.mark.flaky(reruns=5)
def test_straight_2link_2lane_to_3lane():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=3)
    t = 2000
    flow = 2
    W.adddemand("orig", "dest", 0, t, flow)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
        
    assert equal_tolerance(link1.q_mat[2:4,3:5].mean(), 0.8*2)
    assert equal_tolerance(link1.v_mat[2:4,3:5].mean(), 20)
    assert equal_tolerance(link1.k_mat[2:4,3:5].mean(), 0.04*2)
    assert equal_tolerance(link1.q_mat[7:,3:5].mean(), 0.8*2)
    assert equal_tolerance(link1.v_mat[7:,3:5].mean(), 20)
    assert equal_tolerance(link1.k_mat[7:,3:5].mean(), 0.04*2)
    assert equal_tolerance(link2.q_mat[2:,3:5].mean(), 0.8*2)
    assert equal_tolerance(link2.v_mat[2:,3:5].mean(), 20)
    assert equal_tolerance(link2.k_mat[2:,3:5].mean(), 0.04*2)
    
@pytest.mark.flaky(reruns=5)
def test_straight_2link_capacity_out():
        
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    v2s = []

    for i in range(1):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2, capacity_out=1.0)
        link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
        t = 2000
        flow = 1.4
        W.adddemand("orig", "dest", 0, t, flow)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,2:4].mean())
        k1s.append(link1.k_mat[10:,2:4].mean())
        v1s.append(link1.v_mat[10:,2:4].mean())
        q2s.append(link2.q_mat[10:,2:4].mean())
        k2s.append(link2.k_mat[10:,2:4].mean())
        v2s.append(link2.v_mat[10:,2:4].mean())

        
    assert equal_tolerance(np.average(q1s), 1)
    assert equal_tolerance(np.average(k1s), q2k_cong(1, link1))
    assert equal_tolerance(np.average(v1s), 1/q2k_cong(1, link1))
    assert equal_tolerance(np.average(q2s), 1)
    assert equal_tolerance(np.average(k2s), 1/20)
    assert equal_tolerance(np.average(v2s), 20)

@pytest.mark.flaky(reruns=5)
def test_straight_2link_capacity_in():
        
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    v2s = []

    for i in range(1):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2, capacity_in=1.0)
        t = 2000
        flow = 1.4
        W.adddemand("orig", "dest", 0, t, flow)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,2:4].mean())
        k1s.append(link1.k_mat[10:,2:4].mean())
        v1s.append(link1.v_mat[10:,2:4].mean())
        q2s.append(link2.q_mat[10:,2:4].mean())
        k2s.append(link2.k_mat[10:,2:4].mean())
        v2s.append(link2.v_mat[10:,2:4].mean())

        
    assert equal_tolerance(np.average(q1s), 1)
    assert equal_tolerance(np.average(k1s), q2k_cong(1, link1))
    assert equal_tolerance(np.average(v1s), 1/q2k_cong(1, link1))
    assert equal_tolerance(np.average(q2s), 1)
    assert equal_tolerance(np.average(k2s), 1/20)
    assert equal_tolerance(np.average(v2s), 20)


@pytest.mark.flaky(reruns=5)
def test_straight_2link_node_capacity():
        
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    v2s = []

    for i in range(1):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0, flow_capacity=1.0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
        t = 2000
        flow = 1.4
        W.adddemand("orig", "dest", 0, t, flow)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,2:4].mean())
        k1s.append(link1.k_mat[10:,2:4].mean())
        v1s.append(link1.v_mat[10:,2:4].mean())
        q2s.append(link2.q_mat[10:,2:4].mean())
        k2s.append(link2.k_mat[10:,2:4].mean())
        v2s.append(link2.v_mat[10:,2:4].mean())

        
    assert equal_tolerance(np.average(q1s), 1)
    assert equal_tolerance(np.average(k1s), q2k_cong(1, link1))
    assert equal_tolerance(np.average(v1s), 1/q2k_cong(1, link1))
    assert equal_tolerance(np.average(q2s), 1)
    assert equal_tolerance(np.average(k2s), 1/20)
    assert equal_tolerance(np.average(v2s), 20)

@pytest.mark.flaky(reruns=5)
def test_straight_trip_end_in_middle_freeflow():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=3)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)

    W.adddemand("orig", "dest", 0, 2000, 1.6)
    W.adddemand("orig", "mid", 1000, 2000, 0.8)

    W.exec_simulation()
    W.analyzer.compute_edie_state()
    # df = W.analyzer.od_to_pandas()
    # W.analyzer.print_simple_stats()
    # W.analyzer.time_space_diagram_traj_links([link1, link2])
    # W.analyzer.cumulative_curves()
    assert equal_tolerance(link1.q_mat[1:7,3].mean(), 1.6)
    assert equal_tolerance(link1.q_mat[10:,3].mean(), 1.6+0.8)
    assert equal_tolerance(link2.q_mat[1:7,3].mean(), 1.6)
    assert equal_tolerance(link2.q_mat[10:,3].mean(), 1.6)

@pytest.mark.flaky(reruns=5)
def test_straight_trip_end_in_middle():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=3)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
    t = 600
    flow = 1.6
    W.adddemand("orig", "dest", 0, t, flow)
    W.adddemand("orig", "mid", 500, t, flow)

    W.exec_simulation()
    df = W.analyzer.od_to_pandas()
    # W.analyzer.print_simple_stats()
    # W.analyzer.compute_edie_state()
    # W.analyzer.time_space_diagram_traj_links([link1, link2])
    # W.analyzer.cumulative_curves()
    assert equal_tolerance(df["completed_trips"][0], 960)
    assert equal_tolerance(df["completed_trips"][1], 160)
    assert equal_tolerance(df["average_travel_time"][0], 400)
    assert equal_tolerance(df["average_travel_time"][1], 600)

@pytest.mark.flaky(reruns=5)
def test_straight_different_fd():
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    v2s = []

    for i in range(1):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=5, jam_density_per_lane=0.1, number_of_lanes=2)
        t = 2000
        flow = 1.4
        W.adddemand("orig", "dest", 0, t, flow)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,2:4].mean())
        k1s.append(link1.k_mat[10:,2:4].mean())
        v1s.append(link1.v_mat[10:,2:4].mean())
        q2s.append(link2.q_mat[10:,2:4].mean())
        k2s.append(link2.k_mat[10:,2:4].mean())
        v2s.append(link2.v_mat[10:,2:4].mean())

    assert equal_tolerance(np.average(q1s), 0.666)
    assert equal_tolerance(np.average(k1s), q2k_cong(0.666, link1))
    assert equal_tolerance(np.average(v1s), 0.666/q2k_cong(0.666, link1))
    assert equal_tolerance(np.average(q2s), 0.666)
    assert equal_tolerance(np.average(k2s), 0.666/5)
    assert equal_tolerance(np.average(v2s), 5)

@pytest.mark.flaky(reruns=5)
def test_straight_different_fd_different_arg():
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    v2s = []

    for i in range(1):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density=0.4, number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=5, jam_density=0.2, number_of_lanes=2)
        t = 2000
        flow = 1.4
        W.adddemand("orig", "dest", 0, t, flow)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,2:4].mean())
        k1s.append(link1.k_mat[10:,2:4].mean())
        v1s.append(link1.v_mat[10:,2:4].mean())
        q2s.append(link2.q_mat[10:,2:4].mean())
        k2s.append(link2.k_mat[10:,2:4].mean())
        v2s.append(link2.v_mat[10:,2:4].mean())

    assert equal_tolerance(np.average(q1s), 0.666)
    assert equal_tolerance(np.average(k1s), q2k_cong(0.666, link1))
    assert equal_tolerance(np.average(v1s), 0.666/q2k_cong(0.666, link1))
    assert equal_tolerance(np.average(q2s), 0.666)
    assert equal_tolerance(np.average(k2s), 0.666/5)
    assert equal_tolerance(np.average(v2s), 5)


@pytest.mark.flaky(reruns=5)
def test_straight_different_fd_different_lanes():
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    v2s = []

    for i in range(1):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=5, jam_density_per_lane=0.1, number_of_lanes=3)
        t = 2000
        flow = 1.4
        W.adddemand("orig", "dest", 0, t, flow)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,2:4].mean())
        k1s.append(link1.k_mat[10:,2:4].mean())
        v1s.append(link1.v_mat[10:,2:4].mean())
        q2s.append(link2.q_mat[10:,2:4].mean())
        k2s.append(link2.k_mat[10:,2:4].mean())
        v2s.append(link2.v_mat[10:,2:4].mean())
        
    assert equal_tolerance(np.average(q1s), 0.666*1.5)
    assert equal_tolerance(np.average(k1s), q2k_cong(0.666*1.5, link1))
    assert equal_tolerance(np.average(v1s), 0.666*1.5/q2k_cong(0.666*1.5, link1))
    assert equal_tolerance(np.average(q2s), 0.666*1.5)
    assert equal_tolerance(np.average(k2s), 0.666*1.5/5)
    assert equal_tolerance(np.average(v2s), 5)

@pytest.mark.flaky(reruns=5)
def test_straight_different_fd_different_lanes_different_arg():
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    v2s = []

    for i in range(1):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=3000, free_flow_speed=20, jam_density=0.4, number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest", length=3000, free_flow_speed=5, jam_density=0.3, number_of_lanes=3)
        t = 2000
        flow = 1.4
        W.adddemand("orig", "dest", 0, t, flow)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,2:4].mean())
        k1s.append(link1.k_mat[10:,2:4].mean())
        v1s.append(link1.v_mat[10:,2:4].mean())
        q2s.append(link2.q_mat[10:,2:4].mean())
        k2s.append(link2.k_mat[10:,2:4].mean())
        v2s.append(link2.v_mat[10:,2:4].mean())
        
    assert equal_tolerance(np.average(q1s), 0.666*1.5)
    assert equal_tolerance(np.average(k1s), q2k_cong(0.666*1.5, link1))
    assert equal_tolerance(np.average(v1s), 0.666*1.5/q2k_cong(0.666*1.5, link1))
    assert equal_tolerance(np.average(q2s), 0.666*1.5)
    assert equal_tolerance(np.average(k2s), 0.666*1.5/5)
    assert equal_tolerance(np.average(v2s), 5)

@pytest.mark.flaky(reruns=10)
def test_merge_saturated():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig1", 0, 0)
    W.addNode("orig2", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=1)
    link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
    link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=3)
    t1 = 1000
    flow1 = 0.8
    t2 = 1000
    flow2 = 1.6
    W.adddemand("orig1", "dest", 0, t1, flow1)
    W.adddemand("orig2", "dest", 0, t2, flow2)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[3:6,3].mean(), 0.8) 
    assert equal_tolerance(link1.k_mat[3:6,3].mean(), 0.04)
    assert equal_tolerance(link1.v_mat[3:6,3].mean(), 20)
    assert equal_tolerance(link2.q_mat[3:6,3].mean(), 1.6) 
    assert equal_tolerance(link2.k_mat[3:6,3].mean(), 0.08)
    assert equal_tolerance(link2.v_mat[3:6,3].mean(), 20)
    assert equal_tolerance(link3.q_mat[3:6,3].mean(), 2.4) 
    assert equal_tolerance(link3.k_mat[3:6,3].mean(), 0.12)
    assert equal_tolerance(link3.v_mat[3:6,3].mean(), 20)
    

@pytest.mark.flaky(reruns=5)
def test_merge_small_demand():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig1", 0, 0)
    W.addNode("orig2", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest", 2, 1)
    link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=1)
    link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
    link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=3)
    t1 = 1000
    flow1 = 0.8
    t2 = 1000
    flow2 = 0.8
    W.adddemand("orig1", "dest", 0, t1, flow1)
    W.adddemand("orig2", "dest", 0, t2, flow2)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[3:6,3].mean(), 0.8) 
    assert equal_tolerance(link1.k_mat[3:6,3].mean(), 0.04)
    assert equal_tolerance(link1.v_mat[3:6,3].mean(), 20)
    assert equal_tolerance(link2.q_mat[3:6,3].mean(), 0.8) 
    assert equal_tolerance(link2.k_mat[3:6,3].mean(), 0.04)
    assert equal_tolerance(link2.v_mat[3:6,3].mean(), 20)
    assert equal_tolerance(link3.q_mat[3:6,3].mean(), 1.6) 
    assert equal_tolerance(link3.k_mat[3:6,3].mean(), 0.08)
    assert equal_tolerance(link3.v_mat[3:6,3].mean(), 20)

@pytest.mark.flaky(reruns=5)
def test_merge_congested_fair():
    q1s = []
    k1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    v3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=0,
            random_seed=None
        )

        W.addNode("orig1", 0, 0)
        W.addNode("orig2", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        t1 = 2000
        flow1 = 1.4
        t2 = 2000
        flow2 = 1.4
        W.adddemand("orig1", "dest", 0, t1, flow1)
        W.adddemand("orig2", "dest", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[5:,3].mean())
        k1s.append(link1.k_mat[5:,3].mean())
        q2s.append(link2.q_mat[5:,3].mean())
        k2s.append(link2.k_mat[5:,3].mean())
        q3s.append(link3.q_mat[5:,3].mean())
        k3s.append(link3.k_mat[5:,3].mean())
        v3s.append(link3.v_mat[5:,3].mean())

    assert equal_tolerance(np.average(q1s), 0.8)
    assert equal_tolerance(np.average(k1s), q2k_cong(0.8, link1))
    assert equal_tolerance(np.average(q2s), 0.8)
    assert equal_tolerance(np.average(k2s), q2k_cong(0.8, link2))
    assert equal_tolerance(np.average(q3s), 1.6)
    assert equal_tolerance(np.average(k3s), 0.08)

@pytest.mark.flaky(reruns=5)
def test_merge_congested_unfair():
    q1s = []
    k1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    v3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=0,
            random_seed=None
        )

        W.addNode("orig1", 0, 0)
        W.addNode("orig2", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=2, number_of_lanes=2)
        link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        t1 = 2000
        flow1 = 1.4
        t2 = 2000
        flow2 = 1.4
        W.adddemand("orig1", "dest", 0, t1, flow1)
        W.adddemand("orig2", "dest", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[5:,3].mean())
        k1s.append(link1.k_mat[5:,3].mean())
        q2s.append(link2.q_mat[5:,3].mean())
        k2s.append(link2.k_mat[5:,3].mean())
        q3s.append(link3.q_mat[5:,3].mean())
        k3s.append(link3.k_mat[5:,3].mean())
        v3s.append(link3.v_mat[5:,3].mean())

    assert equal_tolerance(np.average(q1s), 1.6*0.666) 
    assert equal_tolerance(np.average(k1s), q2k_cong(1.6*0.666, link1))
    assert equal_tolerance(np.average(q2s), 1.6*0.333)
    assert equal_tolerance(np.average(k2s), q2k_cong(1.6*0.333, link2))
    assert equal_tolerance(np.average(q3s), 1.6) 
    assert equal_tolerance(np.average(k3s), 0.08)
    assert equal_tolerance(np.average(v3s), 20)


@pytest.mark.flaky(reruns=5)
def test_merge_congested_unfair_deltan1():
    q1s = []
    k1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    v3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=1,
            tmax=1000,
            print_mode=0, save_mode=0, show_mode=0,
            random_seed=None
        )

        W.addNode("orig1", 0, 0)
        W.addNode("orig2", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=2, number_of_lanes=2)
        link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        t1 = 1000
        flow1 = 1.4
        t2 = 1000
        flow2 = 1.4
        W.adddemand("orig1", "dest", 0, t1, flow1)
        W.adddemand("orig2", "dest", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[5:,3].mean())
        k1s.append(link1.k_mat[5:,3].mean())
        q2s.append(link2.q_mat[5:,3].mean())
        k2s.append(link2.k_mat[5:,3].mean())
        q3s.append(link3.q_mat[5:,3].mean())
        k3s.append(link3.k_mat[5:,3].mean())
        v3s.append(link3.v_mat[5:,3].mean())

    assert equal_tolerance(np.average(q1s), 1.6*0.666) 
    assert equal_tolerance(np.average(k1s), q2k_cong(1.6*0.666, link1))
    assert equal_tolerance(np.average(q2s), 1.6*0.333)
    assert equal_tolerance(np.average(k2s), q2k_cong(1.6*0.333, link2))
    assert equal_tolerance(np.average(q3s), 1.6) 
    assert equal_tolerance(np.average(k3s), 0.08)
    assert equal_tolerance(np.average(v3s), 20)


@pytest.mark.flaky(reruns=5)
def test_merge_congested_veryunfair():
    q1s = []
    k1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    v3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=0,
            random_seed=None
        )

        W.addNode("orig1", 0, 0)
        W.addNode("orig2", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=10, number_of_lanes=2)
        link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        t1 = 2000
        flow1 = 1.4
        t2 = 2000
        flow2 = 1.4
        W.adddemand("orig1", "dest", 0, t1, flow1)
        W.adddemand("orig2", "dest", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[5:,3].mean())
        k1s.append(link1.k_mat[5:,3].mean())
        q2s.append(link2.q_mat[5:,3].mean())
        k2s.append(link2.k_mat[5:,3].mean())
        q3s.append(link3.q_mat[5:,3].mean())
        k3s.append(link3.k_mat[5:,3].mean())
        v3s.append(link3.v_mat[5:,3].mean())

    assert equal_tolerance(np.average(q1s), 1.4) 
    assert equal_tolerance(np.average(k1s), 1.4/link1.u)
    assert equal_tolerance(np.average(q2s), 0.2)
    assert equal_tolerance(np.average(k2s), q2k_cong(0.2, link2))
    assert equal_tolerance(np.average(q3s), 1.6) 
    assert equal_tolerance(np.average(k3s), 0.08)
    assert equal_tolerance(np.average(v3s), 20)

@pytest.mark.flaky(reruns=5)
def test_merge_congested_different_lanes():
    q1s = []
    k1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    v3s = []
    for i in range(10):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=0,
            random_seed=None
        )

        W.addNode("orig1", 0, 0)
        W.addNode("orig2", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=1)
        link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        t1 = 2000
        flow1 = 0.8
        t2 = 2000
        flow2 = 1.6
        W.adddemand("orig1", "dest", 0, t1, flow1)
        W.adddemand("orig2", "dest", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,6:8].mean())
        k1s.append(link1.k_mat[10:,6:8].mean())
        q2s.append(link2.q_mat[10:,6:8].mean())
        k2s.append(link2.k_mat[10:,6:8].mean())
        q3s.append(link3.q_mat[10:,6:8].mean())
        k3s.append(link3.k_mat[10:,6:8].mean())
        v3s.append(link3.v_mat[10:,6:8].mean())

    #車線の少ないリンクは，車線の多いリンクの試行回数の恩恵を受けて少し有利になる．大きな差はでないので許容する
    assert equal_tolerance(np.average(q1s), 1.6*0.333, rel_tol=0.3) 
    assert equal_tolerance(np.average(k1s), q2k_cong(1.6*0.333, link1), rel_tol=0.3)
    assert equal_tolerance(np.average(q2s), 1.6*0.666, rel_tol=0.3)
    assert equal_tolerance(np.average(k2s), q2k_cong(1.6*0.666, link2), rel_tol=0.3)
    assert equal_tolerance(np.average(q3s), 1.6) 
    assert equal_tolerance(np.average(k3s), 0.08)
    assert equal_tolerance(np.average(v3s), 20)

@pytest.mark.flaky(reruns=5)
def test_merge_congested_signal():
    q1s = []
    k1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    v3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig1", 0, 0)
        W.addNode("orig2", 0, 0)
        W.addNode("mid", 0, 0, signal=[60,60])
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2, signal_group=0)
        link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2, signal_group=1)
        link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        t1 = 2000
        flow1 = 1.6
        t2 = 2000
        flow2 = 1.6
        W.adddemand("orig1", "dest", 0, t1, flow1)
        W.adddemand("orig2", "dest", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,6:8].mean())
        k1s.append(link1.k_mat[10:,6:8].mean())
        q2s.append(link2.q_mat[10:,6:8].mean())
        k2s.append(link2.k_mat[10:,6:8].mean())
        q3s.append(link3.q_mat[10:,6:8].mean())
        k3s.append(link3.k_mat[10:,6:8].mean())
        v3s.append(link3.v_mat[10:,6:8].mean())

    assert equal_tolerance(np.average(q1s), 0.8) 
    assert equal_tolerance(np.average(k1s), q2k_cong(0.8, link1))
    assert equal_tolerance(np.average(q2s), 0.8)
    assert equal_tolerance(np.average(k2s), q2k_cong(0.8, link2))
    assert equal_tolerance(np.average(q3s), 1.6) 
    assert equal_tolerance(np.average(k3s), 0.08)
    assert equal_tolerance(np.average(v3s), 20)

@pytest.mark.flaky(reruns=5)
def test_merge_saturated_signal():
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    v3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig1", 0, 0)
        W.addNode("orig2", 0, 0)
        W.addNode("mid", 0, 0, signal=[60,60])
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2, signal_group=0)
        link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2, signal_group=1)
        link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        t1 = 2000
        flow1 = 0.8
        t2 = 2000
        flow2 = 0.8
        W.adddemand("orig1", "dest", 0, t1, flow1)
        W.adddemand("orig2", "dest", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,6:8].mean())
        k1s.append(link1.k_mat[10:,6:8].mean())
        v1s.append(link1.v_mat[10:,9:].mean())
        q2s.append(link2.q_mat[10:,6:8].mean())
        k2s.append(link2.k_mat[10:,6:8].mean())
        q3s.append(link3.q_mat[10:,6:8].mean())
        k3s.append(link3.k_mat[10:,6:8].mean())
        v3s.append(link3.v_mat[10:,6:8].mean())

    assert equal_tolerance(np.average(q1s), 0.8) 
    assert equal_tolerance(np.average(k1s), 0.8/20)
    assert equal_tolerance(np.average(v1s), 5)
    assert equal_tolerance(np.average(q2s), 0.8)
    assert equal_tolerance(np.average(k2s), 0.8/20)
    assert equal_tolerance(np.average(q3s), 1.6) 
    assert equal_tolerance(np.average(k3s), 0.08)
    assert equal_tolerance(np.average(v3s), 20)

@pytest.mark.flaky(reruns=5)
def test_merge_congested_node_capacity():
    
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    v3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig1", 0, 0)
        W.addNode("orig2", 0, 0)
        W.addNode("mid", 0, 0, flow_capacity=1)
        W.addNode("dest", 2, 1)
        link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=2, number_of_lanes=2)
        link3 = W.addLink("link3", "mid", "dest",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, merge_priority=1, number_of_lanes=2)
        t1 = 2000
        flow1 = 1.6
        t2 = 2000
        flow2 = 1.6
        W.adddemand("orig1", "dest", 0, t1, flow1)
        W.adddemand("orig2", "dest", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[10:,6:8].mean())
        k1s.append(link1.k_mat[10:,6:8].mean())
        v1s.append(link1.v_mat[10:,9:].mean())
        q2s.append(link2.q_mat[10:,6:8].mean())
        k2s.append(link2.k_mat[10:,6:8].mean())
        q3s.append(link3.q_mat[10:,6:8].mean())
        k3s.append(link3.k_mat[10:,6:8].mean())
        v3s.append(link3.v_mat[10:,6:8].mean())

    assert equal_tolerance(np.average(q1s), 0.333) 
    assert equal_tolerance(np.average(k1s), q2k_cong(0.333, link1))
    assert equal_tolerance(np.average(q2s), 0.666)
    assert equal_tolerance(np.average(k2s), q2k_cong(0.666, link2))
    assert equal_tolerance(np.average(q3s), 1.0) 
    assert equal_tolerance(np.average(k3s), 1/20)
    assert equal_tolerance(np.average(v3s), 20)

@pytest.mark.flaky(reruns=5)
def test_merge_disappear_freeflow():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig1", 0, 0)
    W.addNode("orig2", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest1", 2, 1)
    link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
    link3 = W.addLink("link3", "mid", "dest1", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)

    W.adddemand("orig1", "dest1", 0, 2000, 0.8)
    W.adddemand("orig1", "mid",   0, 2000, 0.4)
    W.adddemand("orig2", "dest1", 0, 2000, 0.8)

    W.exec_simulation()
    W.analyzer.print_simple_stats()
    W.analyzer.compute_edie_state()
    #W.analyzer.time_space_diagram_traj()

    assert equal_tolerance(link1.q_mat[3:8,3:5].mean(), 1.2)
    assert equal_tolerance(link2.q_mat[3:8,3:5].mean(), 0.8)
    assert equal_tolerance(link3.q_mat[3:8,3:5].mean(), 1.6)
    assert equal_tolerance(link1.v_mat[3:8,3:5].mean(), 20)
    assert equal_tolerance(link2.v_mat[3:8,3:5].mean(), 20)
    assert equal_tolerance(link3.v_mat[3:8,3:5].mean(), 20)

@pytest.mark.flaky(reruns=5)
def test_merge_disappear_saturated():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig1", 0, 0)
    W.addNode("orig2", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest1", 2, 1)
    link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
    link3 = W.addLink("link3", "mid", "dest1", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)

    W.adddemand("orig1", "dest1", 0, 2000, 0.8)
    W.adddemand("orig1", "mid",   0, 2000, 0.8)
    W.adddemand("orig2", "dest1", 0, 2000, 0.8)

    W.exec_simulation()
    W.analyzer.print_simple_stats()
    W.analyzer.compute_edie_state()

    assert equal_tolerance(link1.q_mat[3:8,3:5].mean(), 1.6)
    assert equal_tolerance(link2.q_mat[3:8,3:5].mean(), 0.8)
    assert equal_tolerance(link3.q_mat[3:8,3:5].mean(), 1.6)
    assert equal_tolerance(link1.v_mat[3:8,3:5].mean(), 20)
    assert equal_tolerance(link2.v_mat[3:8,3:5].mean(), 20)
    assert equal_tolerance(link3.v_mat[3:8,3:5].mean(), 20)

@pytest.mark.flaky(reruns=10)
def test_merge_disappear_congested():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig1", 0, 0)
    W.addNode("orig2", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest1", 2, 1)
    link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
    link3 = W.addLink("link3", "mid", "dest1", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)

    W.adddemand("orig1", "dest1", 0, 2000, 1.2)
    W.adddemand("orig1", "mid",   0, 2000, 0.4)
    W.adddemand("orig2", "dest1", 0, 2000, 0.8)

    W.exec_simulation()
    W.analyzer.print_simple_stats()
    W.analyzer.compute_edie_state()

    assert equal_tolerance(link1.q_mat[3:8,3:5].mean(), 1.3, rel_tol=0.5)
    assert equal_tolerance(link2.q_mat[3:8,3:5].mean(), 0.6, rel_tol=0.5)
    assert equal_tolerance(link3.q_mat[3:8,3:5].mean(), 1.6)
    assert equal_tolerance(link1.v_mat[3:8,3:5].mean(), 9, rel_tol=0.5)
    assert equal_tolerance(link2.v_mat[3:8,3:5].mean(), 8, rel_tol=0.5)
    assert equal_tolerance(link3.v_mat[3:8,3:5].mean(), 20)

@pytest.mark.flaky(reruns=5)
def test_merge_diverge_disappear_freeflow():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig1", 0, 0)
    W.addNode("orig2", 0, 0)
    W.addNode("mid", 0, 0)
    W.addNode("dest1", 2, 1)
    W.addNode("dest2", 0, 0)
    link1 = W.addLink("link1", "orig1", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=3)
    link2 = W.addLink("link2", "orig2", "mid", length=1000, free_flow_speed=30, jam_density_per_lane=0.2, number_of_lanes=2)
    link3 = W.addLink("link3", "mid", "dest1", length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    link4 = W.addLink("link4", "mid", "dest2", length=1000, free_flow_speed=30, jam_density_per_lane=0.2, number_of_lanes=1)

    W.adddemand("orig1", "dest1", 0, 2000, 0.8)
    W.adddemand("orig1", "dest2", 0, 2000, 0.4)
    W.adddemand("orig1", "mid",   0, 2000, 0.8)
    W.adddemand("orig2", "dest1", 0, 2000, 0.8)
    W.adddemand("orig2", "dest2", 0, 2000, 0.4)

    W.exec_simulation()
    W.analyzer.print_simple_stats()
    W.analyzer.compute_edie_state()
    # W.analyzer.time_space_diagram_traj()

    assert equal_tolerance(link1.q_mat[3:8,3:5].mean(), 2.0)
    assert equal_tolerance(link2.q_mat[3:8,3:5].mean(), 1.2)
    assert equal_tolerance(link3.q_mat[3:8,3:5].mean(), 0.8*2)
    assert equal_tolerance(link4.q_mat[3:8,3:5].mean(), 0.8*1)
    assert equal_tolerance(link1.v_mat[3:8,3:5].mean(), 20)
    assert equal_tolerance(link2.v_mat[3:8,3:5].mean(), 30)
    assert equal_tolerance(link3.v_mat[3:8,3:5].mean(), 20)
    assert equal_tolerance(link4.v_mat[3:8,3:5].mean(), 30)

@pytest.mark.flaky(reruns=5)
def test_diverge_saturated_2_2_2():
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=1000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest1", 2, 1)
        W.addNode("dest2", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2,   number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest1", length=1000, free_flow_speed=20, jam_density_per_lane=0.2,  number_of_lanes=2)
        link3 = W.addLink("link3", "mid", "dest2",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
        t1 = 1000
        flow1 = 0.8
        t2 = 1000
        flow2 = 0.8
        W.adddemand("orig", "dest1", 0, t1, flow1)
        W.adddemand("orig", "dest2", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[5:,6:8].mean())
        k1s.append(link1.k_mat[5:,6:8].mean())
        v1s.append(link1.v_mat[5:,6:8].mean())
        q2s.append(link2.q_mat[5:,6:8].mean())
        k2s.append(link2.k_mat[5:,6:8].mean())
        q3s.append(link3.q_mat[5:,6:8].mean())
        k3s.append(link3.k_mat[5:,6:8].mean())

    assert equal_tolerance(np.average(q1s), 1.6) 
    assert equal_tolerance(np.average(k1s), 0.08)
    assert equal_tolerance(np.average(v1s), 20)
    assert equal_tolerance(np.average(q2s), 0.8)
    assert equal_tolerance(np.average(k2s), 0.04)
    assert equal_tolerance(np.average(q3s), 0.8) 
    assert equal_tolerance(np.average(k3s), 0.04)

@pytest.mark.flaky(reruns=5)
def test_diverge_saturated_2_1_1_congestion_due_to_friction(): #TODO: frictionは要精査
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=1000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest1", 2, 1)
        W.addNode("dest2", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2,   number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest1", length=1000, free_flow_speed=20, jam_density_per_lane=0.2,  number_of_lanes=1)
        link3 = W.addLink("link3", "mid", "dest2",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
        t1 = 1000
        flow1 = 0.8
        t2 = 1000
        flow2 = 0.8
        W.adddemand("orig", "dest1", 0, t1, flow1)
        W.adddemand("orig", "dest2", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[5:,6:8].mean())
        k1s.append(link1.k_mat[5:,6:8].mean())
        v1s.append(link1.v_mat[5:,6:8].mean())
        q2s.append(link2.q_mat[5:,6:8].mean())
        k2s.append(link2.k_mat[5:,6:8].mean())
        q3s.append(link3.q_mat[5:,6:8].mean())
        k3s.append(link3.k_mat[5:,6:8].mean())

    assert equal_tolerance(np.average(q1s), 1.44) 
    assert equal_tolerance(np.average(k1s), 0.11)
    assert equal_tolerance(np.average(v1s), 13)
    assert equal_tolerance(np.average(q2s), 0.72)
    assert equal_tolerance(np.average(k2s), 0.036)
    assert equal_tolerance(np.average(q3s), 0.72) 
    assert equal_tolerance(np.average(k3s), 0.036)

@pytest.mark.flaky(reruns=10)
def test_diverge_lesssaturated_2_1_1():
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=1000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest1", 2, 1)
        W.addNode("dest2", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2,   number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest1", length=1000, free_flow_speed=20, jam_density_per_lane=0.2,  number_of_lanes=1)
        link3 = W.addLink("link3", "mid", "dest2",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
        t1 = 1000
        flow1 = 0.7
        t2 = 1000
        flow2 = 0.7
        W.adddemand("orig", "dest1", 0, t1, flow1)
        W.adddemand("orig", "dest2", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[5:,6:8].mean())
        k1s.append(link1.k_mat[5:,6:8].mean())
        v1s.append(link1.v_mat[5:,6:8].mean())
        q2s.append(link2.q_mat[5:,6:8].mean())
        k2s.append(link2.k_mat[5:,6:8].mean())
        q3s.append(link3.q_mat[5:,6:8].mean())
        k3s.append(link3.k_mat[5:,6:8].mean())

    assert equal_tolerance(np.average(q1s), 1.4) 
    assert equal_tolerance(np.average(k1s), 0.07)
    assert equal_tolerance(np.average(v1s), 20)
    assert equal_tolerance(np.average(q2s), 0.7)
    assert equal_tolerance(np.average(k2s), 0.035)
    assert equal_tolerance(np.average(q3s), 0.7) 
    assert equal_tolerance(np.average(k3s), 0.035)

@pytest.mark.flaky(reruns=5)
def test_diverge_congested_2_1_1():
    q1s = []
    k1s = []
    v1s = []
    q2s = []
    k2s = []
    q3s = []
    k3s = []
    for i in range(5):
        W = World(
            name="",
            deltan=5,
            tmax=1000,
            print_mode=0, save_mode=0, show_mode=1,
            random_seed=None
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid", 0, 0)
        W.addNode("dest1", 2, 1)
        W.addNode("dest2", 2, 1)
        link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density_per_lane=0.2,   number_of_lanes=2)
        link2 = W.addLink("link2", "mid", "dest1", length=1000, free_flow_speed=20, jam_density_per_lane=0.2,  number_of_lanes=1)
        link3 = W.addLink("link3", "mid", "dest2",  length=1000, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=1)
        t1 = 1000
        flow1 = 1.6*2/3
        t2 = 1000
        flow2 = 1.6*1/3
        W.adddemand("orig", "dest1", 0, t1, flow1)
        W.adddemand("orig", "dest2", 0, t2, flow2)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        W.analyzer.compute_edie_state()
        q1s.append(link1.q_mat[5:,6:8].mean())
        k1s.append(link1.k_mat[5:,6:8].mean())
        v1s.append(link1.v_mat[5:,6:8].mean())
        q2s.append(link2.q_mat[5:,6:8].mean())
        k2s.append(link2.k_mat[5:,6:8].mean())
        q3s.append(link3.q_mat[5:,6:8].mean())
        k3s.append(link3.k_mat[5:,6:8].mean())

    assert equal_tolerance(np.average(q1s), 1.2) 
    assert equal_tolerance(np.average(k1s), q2k_cong(1.2, link1))
    assert equal_tolerance(np.average(v1s), 1.2/q2k_cong(1.2, link1))
    assert equal_tolerance(np.average(q2s), 0.8)
    assert equal_tolerance(np.average(k2s), 0.04)
    assert equal_tolerance(np.average(q3s), 0.4) 
    assert equal_tolerance(np.average(k3s), 0.4/20)

@pytest.mark.flaky(reruns=10)
def test_4phase_signal_jpstyle_straight_2lane():
    W = World(
        name="",
        deltan=1,
        tmax=1500,
        print_mode=1, save_mode=0, show_mode=0,
        random_seed=None
    )

    # シナリオ定義
    #4現示信号:
    #    現示0: 南北直左のみ50秒
    #    現示1: 南北右折のみ10秒
    #    現示2: 東西直左のみ40秒
    #    現示3: 東西右折のみ5秒
    W.addNode("S_orig", 0, -3)    #起点ノード
    W.addNode("N_orig", 0, 3)
    W.addNode("W_orig", -3, 0)
    W.addNode("E_orig", 3, 0)
    W.addNode("S_dest", 0, -3)    #終点ノード
    W.addNode("N_dest", 0, 3)
    W.addNode("W_dest", -3, 0)
    W.addNode("E_dest", 3, 0)

    signal_phase_setting = [50,10,40,5]
    W.addNode("S_i", 0, -1, signal=signal_phase_setting)  #信号交差点の南側入口ノード
    W.addNode("N_i", 0.2, 1, signal=signal_phase_setting)
    W.addNode("W_i", -1, 0.2, signal=signal_phase_setting)
    W.addNode("E_i", 1, 0, signal=signal_phase_setting)
    W.addNode("S_o", 0.2, -1, signal=signal_phase_setting)  #信号交差点の南側出口ノード
    W.addNode("N_o", 0, 1, signal=signal_phase_setting)
    W.addNode("W_o", -1, 0, signal=signal_phase_setting)
    W.addNode("E_o", 1, 0.2, signal=signal_phase_setting)

    W.addLink("linkSin", "S_orig", "S_i", length=200, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=[0,1], number_of_lanes=2)   #交差点への流入リンク
    W.addLink("linkNin", "N_orig", "N_i", length=200, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=[0,1], number_of_lanes=2)
    W.addLink("linkWin", "W_orig", "W_i", length=200, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=[2,3], number_of_lanes=2)
    W.addLink("linkEin", "E_orig", "E_i", length=200, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=[2,3], number_of_lanes=2)
    W.addLink("linkSout", "S_o", "S_dest", length=200, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)   #交差点からの流出リンク
    W.addLink("linkNout", "N_o", "N_dest", length=200, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    W.addLink("linkWout", "W_o", "W_dest", length=200, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)
    W.addLink("linkEout", "E_o", "E_dest", length=200, free_flow_speed=20, jam_density_per_lane=0.2, number_of_lanes=2)

    #交差点内部ダミーリンク
    W.addLink("signal_SN_s", "S_i", "N_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=0, number_of_lanes=2)  #直進．lengthは仮のもの．あまりに短すぎると丸め誤差で変なことがおきるかも
    W.addLink("signal_NS_s", "N_i", "S_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=0, number_of_lanes=2)
    W.addLink("signal_SW_l", "S_i", "W_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=0)  #左折
    W.addLink("signal_NE_l", "N_i", "E_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=0)
    W.addLink("signal_SE_r", "S_i", "E_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=1)  #右折．lengthは右折待ち行列の最大長さに相当
    W.addLink("signal_NW_r", "N_i", "W_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=1)
    W.addLink("signal_WE_s", "W_i", "E_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=2, number_of_lanes=2)
    W.addLink("signal_EW_s", "E_i", "W_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=2, number_of_lanes=2)
    W.addLink("signal_WN_l", "W_i", "N_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=2)
    W.addLink("signal_ES_l", "E_i", "S_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=2)
    W.addLink("signal_WS_r", "W_i", "S_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=3)
    W.addLink("signal_EN_r", "E_i", "N_o", length=20, free_flow_speed=20, jam_density_per_lane=0.2, signal_group=3)

    demand_s = 0.4
    demand_l = 0.03
    demand_r = 0.03

    duration = 1000
    dt = 50
    randomness = (0.7, 1.3)
    for t in range(0, duration+1, dt):
        W.adddemand("S_orig", "N_dest", t, t+dt, demand_s*random.uniform(*randomness))
        W.adddemand("S_orig", "E_dest", t, t+dt, demand_l*random.uniform(*randomness))
        W.adddemand("S_orig", "W_dest", t, t+dt, demand_r*random.uniform(*randomness))
        W.adddemand("N_orig", "S_dest", t, t+dt, demand_s*random.uniform(*randomness))
        W.adddemand("N_orig", "E_dest", t, t+dt, demand_l*random.uniform(*randomness))
        W.adddemand("N_orig", "W_dest", t, t+dt, demand_r*random.uniform(*randomness))
        W.adddemand("W_orig", "E_dest", t, t+dt, demand_s*random.uniform(*randomness))
        W.adddemand("W_orig", "N_dest", t, t+dt, demand_l*random.uniform(*randomness))
        W.adddemand("W_orig", "S_dest", t, t+dt, demand_r*random.uniform(*randomness))
        W.adddemand("E_orig", "W_dest", t, t+dt, demand_s*random.uniform(*randomness))
        W.adddemand("E_orig", "N_dest", t, t+dt, demand_l*random.uniform(*randomness))
        W.adddemand("E_orig", "S_dest", t, t+dt, demand_r*random.uniform(*randomness))

    # シミュレーション実行
    #最後までシミュを回す
    W.exec_simulation()

    df = W.analyzer.od_to_pandas()

    # 結果可視化
    W.analyzer.print_simple_stats()

    print("南から北へ直進する経路")
    W.analyzer.time_space_diagram_traj_links(["linkSin","signal_SN_s", "linkNout"], xlim=[0,300])
    W.analyzer.time_space_diagram_traj_links(["linkSin","signal_SN_s", "linkNout"])
    print("南から西へ左折する経路")
    W.analyzer.time_space_diagram_traj_links(["linkSin","signal_SW_l", "linkWout"], xlim=[0,300])
    W.analyzer.time_space_diagram_traj_links(["linkSin","signal_SW_l", "linkWout"])
    print("南から東へ右折する経路")
    W.analyzer.time_space_diagram_traj_links(["linkSin","signal_SE_r", "linkEout"], xlim=[0,300])
    W.analyzer.time_space_diagram_traj_links(["linkSin","signal_SE_r", "linkEout"])

    avt = df["average_travel_time"].values
    print(avt)
    
    referemce_avt = [41.40449658, 65.77142857, 41.56507937, 41.41644018, 41.30793651, 65.97619048, 49.7849498, 49.15555556, 69.54444444, 49.11051701, 70.71428571, 48.62857143]
    for i in range(len(avt)):
        assert equal_tolerance(avt[i], referemce_avt[i], rel_tol=0.2)

@pytest.mark.flaky(reruns=5)
def test_route_choice_one_is_too_long_and_another_has_bottleneck():
    vol1s = []
    vol2s = []
    atts1 = []
    atts2 = []

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=3000, 
            print_mode=0, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2, number_of_lanes=2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2, number_of_lanes=1)
        link21 = W.addLink("link21", "orig", "mid2", length=2000, free_flow_speed=20, jam_density=0.2, number_of_lanes=2)
        link22 = W.addLink("link22", "mid2", "dest", length=2000, free_flow_speed=20, jam_density=0.2, number_of_lanes=2)
        W.adddemand("orig", "dest", 0, 3000, 1.6)

        W.exec_simulation()
        

        W.analyzer.print_simple_stats()

        W.analyzer.basic_analysis()

        W.analyzer.compute_edie_state()
        # W.analyzer.time_space_diagram_traj_links([link11, link12])
        # W.analyzer.time_space_diagram_traj_links([link21, link22])

        vol1s.append(link11.q_mat[5:,5].sum()*link11.edie_dt)
        vol2s.append(link21.q_mat[5:,5].sum()*link21.edie_dt)
        for veh in W.VEHICLES.values():
            if link12 in veh.log_link:
                atts1.append(veh.travel_time)
            elif link22 in veh.log_link:
                atts2.append(veh.travel_time)
        
    # plt.hist(atts1, bins=100, color="red", alpha=0.5)
    # plt.hist(atts2, bins=100, color="blue", alpha=0.5)
    print(f"{np.average(vol1s) = }\n{np.average(vol2s) = }\n{np.average(atts1) = }\n{np.average(atts2) = }")
    assert equal_tolerance(np.average(vol1s), 1915)
    assert equal_tolerance(np.average(vol1s), 1915)
    assert equal_tolerance(np.average(atts1), 200)
    assert equal_tolerance(np.average(atts2), 200)

def test_route_choice_4route_congestion_avoidance():
    tt1s = []
    tt2s = []
    tt3s = []
    tt4s = []
    vol1s = []
    vol2s = []
    vol3s = []
    vol4s = []
    ttas = []

    for i in range(20):
        W = World(
            name="",
            deltan=5, 
            tmax=3000, 
            print_mode=0, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100,
            duo_update_weight=0.3
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", 2, 1)
        W.addNode("mid3", 3, 1)
        W.addNode("mid4", 4, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.4, number_of_lanes=2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2, number_of_lanes=1)
        link21 = W.addLink("link21", "orig", "mid2", length=1000, free_flow_speed=20, jam_density=0.4, number_of_lanes=2)
        link22 = W.addLink("link22", "mid2", "dest", length=1000, free_flow_speed=20, jam_density=0.2, number_of_lanes=1)
        link31 = W.addLink("link31", "orig", "mid3", length=1000, free_flow_speed=20, jam_density=0.4, number_of_lanes=2)
        link32 = W.addLink("link32", "mid3", "dest", length=1000, free_flow_speed=20, jam_density=0.2, number_of_lanes=1)
        link41 = W.addLink("link41", "orig", "mid4", length=1000, free_flow_speed=20, jam_density=0.4, number_of_lanes=2)
        link42 = W.addLink("link42", "mid4", "dest", length=1000, free_flow_speed=20, jam_density=0.2, number_of_lanes=1)
        W.adddemand("orig", "dest", 0, 2000, 0.8*4)

        W.exec_simulation() 

        W.analyzer.print_simple_stats()

        # W.analyzer.time_space_diagram_traj_links([link11, link12])
        # W.analyzer.time_space_diagram_traj_links([link21, link22])
        # W.analyzer.time_space_diagram_traj_links([link31, link32])
        # W.analyzer.time_space_diagram_traj_links([link41, link42])
        W.analyzer.basic_analysis()

        df = W.analyzer.link_to_pandas()

        tt1 = df[df["link"].isin(("link11", "link12"))]["average_travel_time"].sum()
        tt2 = df[df["link"].isin(("link21", "link22"))]["average_travel_time"].sum()
        tt3 = df[df["link"].isin(("link31", "link32"))]["average_travel_time"].sum()
        tt4 = df[df["link"].isin(("link41", "link42"))]["average_travel_time"].sum()
        vol1 = df[df["link"].isin(["link11"])]["traffic_volume"].values[0]
        vol2 = df[df["link"].isin(["link21"])]["traffic_volume"].values[0]
        vol3 = df[df["link"].isin(["link31"])]["traffic_volume"].values[0]
        vol4 = df[df["link"].isin(["link41"])]["traffic_volume"].values[0]

        df2 = W.analyzer.od_to_pandas()
        tta = df2["average_travel_time"].values[0]

        tt1s.append(tt1)
        tt2s.append(tt2)
        tt3s.append(tt3)
        tt4s.append(tt4)
        vol1s.append(vol1)
        vol2s.append(vol2)
        vol3s.append(vol3)
        vol4s.append(vol4)
            
        ttas.append(tta)

    print(f"{np.average(tt1s) = }\n{np.average(tt2s) = }\n{np.average(tt3s) = }\n{np.average(tt4s) = }\n{np.average(vol1s) = }\n{np.average(vol2s) = }\n{np.average(vol3s) = }\n{np.average(vol4s) = }\n{np.average(np.concatenate((tt1s, tt2s, tt3s, tt4s))) = }\n{np.average(np.concatenate((vol1s, vol2s, vol3s, vol4s))) = }")
    ttave = np.average(np.concatenate((tt1s, tt2s, tt3s, tt4s)))
    volave = np.average(np.concatenate((vol1s, vol2s, vol3s, vol4s)))

    assert equal_tolerance(np.average(tt1s), ttave, rel_tol=0.2)
    assert equal_tolerance(np.average(tt2s), ttave, rel_tol=0.2)
    assert equal_tolerance(np.average(tt3s), ttave, rel_tol=0.2)
    assert equal_tolerance(np.average(tt4s), ttave, rel_tol=0.2)
    assert equal_tolerance(np.average(vol1s), volave, rel_tol=0.2)
    assert equal_tolerance(np.average(vol2s), volave, rel_tol=0.2)
    assert equal_tolerance(np.average(vol3s), volave, rel_tol=0.2)
    assert equal_tolerance(np.average(vol4s), volave, rel_tol=0.2)
    assert equal_tolerance(ttave, 225, rel_tol=0.2)
    assert equal_tolerance(volave*4, 2000*0.8*4, rel_tol=0.2)



"""
This script test Jupyter notebooks that are not tested by `test_notebook_demos.py`. Since it takes long time, this test is excluded from the merging checker.
Implementation is awkward, but it works. 
"""


import pytest
import nbformat
import os
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors import CellExecutionError

from IPython.display import display
import re


class CellExecuter(ExecutePreprocessor):
    """特定のキーワードを含むセルをスキップするプリプロセッサ"""
    
    def __init__(self, exception_words=None, **kwargs):
        super().__init__(**kwargs)
        self.exception_words = exception_words or []
    
    def preprocess_cell(self, cell, resources, cell_index):
        """セルを処理する前に、スキップすべきかどうかをチェック"""
        if cell.cell_type == "code":
            if any(word in cell.source for word in self.exception_words):
                print(f"Skipping cell {cell_index}: {cell.source[:50]}...")
                return cell, resources
        
        return super().preprocess_cell(cell, resources, cell_index)
    

def test_demo_notebook_02en():
    # スキップするワードのリスト
    exception_words = ["ResultGUIViewer", "%matplotlib"]
    notebook_dir = "demos_and_examples"
    
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_02en" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = CellExecuter(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")

def test_demo_notebook_03en_pytorch():
    exception_words = [ "ResultGUIViewer", "%matplotlib", "%run"]
    notebook_dir = "demos_and_examples"
    
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_03en_pytorch" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = CellExecuter(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")

def test_demo_notebook_04en_OpenStreetMap():
    exception_words = [ "ResultGUIViewer", "%matplotlib", "osm_tokyo.png"]

    notebook_dir = "demos_and_examples"
    
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_04en_OpenStreetMap" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = CellExecuter(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")

def test_demo_notebook_06en_taxi_or_shared_mobility():
    exception_words = [ "ResultGUIViewer", "%matplotlib"]

    notebook_dir = "demos_and_examples"
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_06en_taxi_or_shared_mobility" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = CellExecuter(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")


def test_demo_notebook_08en_chicago():
    exception_words = [ "ResultGUIViewer", "%matplotlib", "df_tntp_flow", "res_tntp", "res_tntp_sec*res_tntp_vol", "tntp parser customized for Chicago-sketch"]

    notebook_dir = "demos_and_examples"
    
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_08en_chicago" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = CellExecuter(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")

def test_demo_notebook_09en_dynamic_traffic_assignment():
    exception_words = [ "ResultGUIViewer", "%matplotlib"]

    notebook_dir = "demos_and_examples"
    
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_09en_dynamic_traffic_assignment" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = CellExecuter(
                exception_words=exception_words,
                timeout=7200, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")



def test_demo_notebook_10en_traffic_signal_tutorial():
    exception_words = [ "ResultGUIViewer", "%matplotlib", "df_tntp_flow", "res_tntp", "res_tntp_sec*res_tntp_vol", "tntp parser customized for Chicago-sketch"]

    notebook_dir = "demos_and_examples"
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_10en_traffic_signal_tutorial" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = CellExecuter(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")


"""
Adaptive Large Neighborhood Search (ALNS) for DSO

- Initially coded by GPT-5, refined by human and Claude Code. 
- Customizable minimum implementation
- Includes domain-knowledge-based heuristics
"""

from __future__ import annotations
from dataclasses import dataclass, field, asdict
from typing import Callable, Sequence, List, Optional, Dict, Any, Tuple, Iterable
import math
import random
import warnings
import numpy as np
from pprint import pprint
import time

from ..Utilities import estimate_congestion_externality_route

# -------------------------
# ログ構造体
# -------------------------
@dataclass
class StepLog:
    """ALNSステップのログ情報.
    
    Attributes
    ----------
    iter : int
        イテレーション番号
    move : str
        実行されたムーブ ("D:<destroy>+R:<repair>")
    changed_idx : Optional[List[int]]
        変更されたインデックスのリスト
    changed_idxs : Optional[int]
        変更されたインデックスの数
    delta_obj : float
        内部最小化差分 (cand - cur)
    accepted : bool
        ムーブが受理されたかどうか
    obj_val : float
        ユーザー向き目的関数値（sense反映）
    best_obj : float
        ユーザー向き最良目的関数値
    temperature : float
        現在の温度
    source : str
        ステップのソース ("auto")
    """
    iter: int
    move: str                          # "D:<destroy>+R:<repair>"
    changed_idx: Optional[List[int]]
    changed_idxs: Optional[int]
    delta_obj: float                   # 内部最小化差分 cand - cur
    accepted: bool
    obj_val: float                     # ユーザー向き（sense 反映）
    best_obj: float                    # ユーザー向き
    temperature: float
    source: str                        # "auto"

@dataclass
class RunLog:
    """ALNS実行全体のログ情報.
    
    Attributes
    ----------
    steps : List[StepLog]
        各ステップのログリスト
    best_x : Sequence[Any]
        最良解
    best_obj : float
        最良目的関数値
    seed : Optional[int]
        使用された乱数シード
    stop_reason : str
        停止理由
    operator_stats : Dict[str, Any]
        オペレータの統計情報 {"destroy":..., "repair":...}
    params : Dict[str, Any]
        使用されたパラメータ
    
    Methods
    -------
    to_dict()
        辞書形式に変換
    to_dataframe()
        pandasデータフレームに変換（pandasが必要）
    """
    steps: List[StepLog] = field(default_factory=list)
    best_x: Sequence[Any] = field(default_factory=list)
    best_obj: float = float("inf")
    seed: Optional[int] = None
    stop_reason: str = ""
    operator_stats: Dict[str, Any] = field(default_factory=dict)   # {"destroy":..., "repair":...}
    params: Dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def to_dataframe(self):
        try:
            import pandas as pd
            return pd.DataFrame([s.__dict__ for s in self.steps])
        except Exception:
            return [s.__dict__ for s in self.steps]

# -------------------------
# 内部ユーティリティ
# -------------------------
def _copy_vec(x: Sequence[Any]) -> List[Any]:
    """ベクトルをコピー.
    
    numpy配列の場合はリストに変換してコピー。
    
    Parameters
    ----------
    x : Sequence[Any]
        コピー元のベクトル
    
    Returns
    -------
    List[Any]
        コピーされたリスト
    """
    if isinstance(x, np.ndarray):
        return x.copy().tolist()
    return list(x)

def _rng_choice(rng: random.Random, seq: Sequence[Any]) -> Any:
    """乱数生成器を使ってシーケンスから要素をランダム選択.
    
    Parameters
    ----------
    rng : random.Random
        乱数生成器
    seq : Sequence[Any]
        選択元のシーケンス
    
    Returns
    -------
    Any
        選択された要素
    """
    return seq[rng.randrange(len(seq))]

def _infer_domains_from_x(x: Sequence[Any]) -> List[List[Any]]:
    """解ベクトルから定義域を自動推定.
    
    各位置の値として、xに含まれる全てのユニーク値を
    使用可能と仮定。
    
    Parameters
    ----------
    x : Sequence[Any]
        推定元の解ベクトル
    
    Returns
    -------
    List[List[Any]]
        各位置の定義域
    """
    uniq = sorted(set(x), key=lambda v: (str(type(v)), str(v)))
    return [uniq[:] for _ in range(len(x))]

# -------------------------
# 状態
# -------------------------
@dataclass
class ALNSState:
    """ALNS実行状態.
    
    アルゴリズムの全状態を保持するデータクラス。
    
    Attributes
    ----------
    sense : str
        最適化の方向 ("min" または "max")
    obj : Callable[[Sequence[Any]], float]
        目的関数
    domains : List[Sequence[Any]]
        各変数の定義域
    destroy_set : List[str]
        使用する破壊オペレータの名前リスト
    repair_set : List[str]
        使用する修復オペレータの名前リスト
    k_min : int
        破壊規模の最小値
    k_max : int
        破壊規模の最大値
    k_cur : int
        破壊規模の現在値
    T0 : float
        初期温度
    Tend : float
        終了温度
    total_iters : int
        総イテレーション数
    segment_len : int
        適応更新のセグメント長
    adapt_eta : float
        適応更新の学習率
    reward_improve_best : float
        最良解改善時の報酬
    reward_improve : float
        解改善時の報酬
    reward_accept_worse : float
        悪化解受理時の報酬
    rng : random.Random
        乱数生成器
    seed : Optional[int]
        乱数シード
    params : Dict[str, Any]
        パラメータ辞書
    additional_info : Dict[str, Any]
        追加情報
    x_cur : List[Any]
        現在解
    x_best : List[Any]
        最良解
    cur : float
        現在解の目的関数値（内部用）
    best : float
        最良解の目的関数値（内部用）
    it : int
        現在のイテレーション番号
    last_best_it : int
        最後に最良解が更新されたイテレーション番号
    d_weights : Dict[str, float]
        破壊オペレータの重み
    d_scores : Dict[str, float]
        破壊オペレータのスコア
    d_uses : Dict[str, int]
        破壊オペレータの使用回数
    d_best_improves : Dict[str, int]
        破壊オペレータによる最良解改善回数
    r_weights : Dict[str, float]
        修復オペレータの重み
    r_scores : Dict[str, float]
        修復オペレータのスコア
    r_uses : Dict[str, int]
        修復オペレータの使用回数
    r_best_improves : Dict[str, int]
        修復オペレータによる最良解改善回数
    seg_countdown : int
        次のセグメント更新までのカウントダウン
    steps : List[StepLog]
        ステップログのリスト
    """
    # パラメータ
    sense: str
    obj: Callable[[Sequence[Any]], float]
    domains: List[Sequence[Any]]

    # ---- 破壊/修復（可変オペレータ集合） ----
    destroy_set: List[str]
    repair_set: List[str]

    # 破壊規模
    k_min: int
    k_max: int
    k_cur: int

    # 焼きなまし
    T0: float
    Tend: float
    total_iters: int

    # 適応（セグメント更新）
    segment_len: int
    adapt_eta: float
    reward_improve_best: float
    reward_improve: float
    reward_accept_worse: float
    always_accept: bool


    rng: random.Random
    seed: Optional[int]
    params: Dict[str, Any]
    additional_info: Dict[str, Any]

    # 状態
    x_cur: List[Any]
    x_best: List[Any]
    cur: float
    best: float
    it: int = 0
    last_best_it: int = 0

    # 破壊/修復の適応重み
    d_weights: Dict[str, float] = field(default_factory=dict)
    d_scores: Dict[str, float] = field(default_factory=dict)
    d_uses: Dict[str, int] = field(default_factory=dict)
    d_best_improves: Dict[str, int] = field(default_factory=dict)

    r_weights: Dict[str, float] = field(default_factory=dict)
    r_scores: Dict[str, float] = field(default_factory=dict)
    r_uses: Dict[str, int] = field(default_factory=dict)
    r_best_improves: Dict[str, int] = field(default_factory=dict)

    seg_countdown: int = 0

    # ログ
    steps: List[StepLog] = field(default_factory=list)

# -------------------------
# 破壊オペレータ実装
#   シグネチャ: (state, base) -> (removed_idx: list[int], partial_x: list[Any] with None on removed)
# -------------------------

def destroy_random(state: ALNSState, base: Sequence[Any]) -> Tuple[List[int], List[Any]]:
    """ランダムにk個の要素を選んで破壊.
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    base : Sequence[Any]
        基本となる解
    
    Returns
    -------
    Tuple[List[int], List[Any]]
        (削除されたインデックス, 部分解) のタプル
        部分解の削除位置はNoneに設定される
    """
    n = len(base)
    k = state.rng.randint(state.k_min, min(state.k_cur, n))
    removed = state.rng.sample(range(n), k)
    xx = _copy_vec(base)
    for i in removed:
        xx[i] = None
    return removed, xx

def destroy_segment(state: ALNSState, base: Sequence[Any]) -> Tuple[List[int], List[Any]]:
    """連続したセグメントを破壊.
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    base : Sequence[Any]
        基本となる解
    
    Returns
    -------
    Tuple[List[int], List[Any]]
        (削除されたインデックス, 部分解) のタプル
        部分解の削除位置はNoneに設定される
    """
    n = len(base)
    if n < 2:
        return destroy_random(state, base)
    i = state.rng.randrange(n)
    j = state.rng.randrange(n)
    if i > j:
        i, j = j, i
    length = j - i + 1
    if length < state.k_min:
        j = min(n - 1, i + state.k_min - 1)
    if (j - i + 1) > state.k_cur:
        j = i + state.k_cur - 1
    removed = list(range(i, j + 1))
    xx = _copy_vec(base)
    for t in removed:
        xx[t] = None
    return removed, xx

def destroy_early_departure(state: ALNSState, base: Sequence[Any]) -> Tuple[List[int], List[Any]]:
    """出発時刻の早い車両を優先で重み付き確率でk個の要素を選んで破壊.
        
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    base : Sequence[Any]
        基本となる解
    
    Returns
    -------
    Tuple[List[int], List[Any]]
        (削除されたインデックス, 部分解) のタプル
        部分解の削除位置はNoneに設定される
    """
    n = len(base)
    
    #出発時刻の早いものを選択しやすい
    w = (1 - state.additional_info.get("departure_times"))
    w = np.maximum(1/100000, w-0.3)
    
    k = state.rng.randint(state.k_min, min(state.k_cur, n))
    probs = (w / w.sum())
    removed = list(np.random.choice(np.arange(n), size=k, replace=False, p=probs))
    xx = _copy_vec(base)
    for i in removed:
        xx[i] = None
    return removed, xx

def destroy_late_departure(state: ALNSState, base: Sequence[Any]) -> Tuple[List[int], List[Any]]:
    """出発時刻の遅い車両を優先で重み付き確率でk個の要素を選んで破壊.
        
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    base : Sequence[Any]
        基本となる解
    
    Returns
    -------
    Tuple[List[int], List[Any]]
        (削除されたインデックス, 部分解) のタプル
        部分解の削除位置はNoneに設定される
    """
    n = len(base)
    
    #出発時刻の遅いものを選択しやすい
    w = state.additional_info.get("departure_times")
    w = np.maximum(1/100000, w-0.3)
    
    k = state.rng.randint(state.k_min, min(state.k_cur, n))
    probs = (w / w.sum())
    removed = list(np.random.choice(np.arange(n), size=k, replace=False, p=probs))
    xx = _copy_vec(base)
    for i in removed:
        xx[i] = None
    return removed, xx


def destroy_congested_link(state: ALNSState, base: Sequence[Any]) -> Tuple[List[int], List[Any]]:
    """混雑リンクを走行した車両を選んで破壊
        
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    base : Sequence[Any]
        基本となる解
    
    Returns
    -------
    Tuple[List[int], List[Any]]
        (削除されたインデックス, 部分解) のタプル
        部分解の削除位置はNoneに設定される
    """
    n = len(base)

    removed = []

    #混雑リンクを選択
    W = state.additional_info.get("W")
    df_l = W.analyzer.link_to_pandas()
    
    delay_ranking_ratio = 0.05
    delay_ranking_min_n = 3

    most_delayed_links = list(df_l.sort_values(by='delay_ratio', ascending=False)[:max([int(len(df_l)*delay_ranking_ratio),delay_ranking_min_n])]["link"])
    
    k = state.rng.randint(state.k_min, min(state.k_cur, n))
    for _ in range(k):
        link_selected_name = np.random.choice(most_delayed_links)
        link_selected = W.get_link(link_selected_name)

        veh_selected = np.random.choice(list(link_selected.vehicles_enter_log.values()))
        removed.append(veh_selected.id)

    xx = _copy_vec(base)
    for i in removed:
        xx[i] = None
    return removed, xx


# 破壊オペレータ一覧
_DESTROY_IMPLS: Dict[str, Callable[[ALNSState, Sequence[Any]], Tuple[List[int], List[Any]]]] = {
    "random": destroy_random,
    "segment": destroy_segment,
    "early_dep": destroy_early_departure,
    "late_dep": destroy_late_departure,
    "congest_link": destroy_congested_link,
}

# 破壊オペレータの初期重み（選択確率に比例）
_DESTROY_INITIAL_WEIGHTS: Dict[str, float] = {
    "random": 1.0,
    "segment": 1.0,
    "early_dep": 5.0,
    "late_dep": 0.2,
    "congest_link": 1.0,
}

# -------------------------
# 修復オペレータ実装
#   シグネチャ: (state, base_cur, partial_x, removed_idx) -> (x_filled: list[Any], changed_idx: list[int])
# -------------------------
def _eval_internal(state: ALNSState, x: Sequence[Any]) -> float:
    """内部用の目的関数評価.
    
    最大化問題の場合は符号を反転して最小化として扱う。
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    x : Sequence[Any]
        評価する解
    
    Returns
    -------
    float
        内部目的関数値（常に最小化）
    """
    v = state.obj(x)
    return -v if state.sense == "max" else v

def _repair_random(state: ALNSState, base_cur: Sequence[Any], partial_x: Sequence[Any], removed: List[int]) -> Tuple[List[Any], List[int]]:
    """ランダムに値を選んで修復.
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    base_cur : Sequence[Any]
        現在の基本解
    partial_x : Sequence[Any]
        部分解（削除位置はNone）
    removed : List[int]
        削除されたインデックス
    
    Returns
    -------
    Tuple[List[Any], List[int]]
        (修復された解, 変更されたインデックス) のタプル
    """
    x = _copy_vec(partial_x)
    rng = state.rng
    for i in removed:
        cur_val = base_cur[i]
        choices = [c for c in state.domains[i]]
        if len(choices) > 1:
            choices = [c for c in choices if c != cur_val] or choices
        x[i] = _rng_choice(rng, choices)
    return x, removed[:]

def _repair_minimum_cost(state: ALNSState, base_cur: Sequence[Any], partial_x: Sequence[Any], removed: List[int]) -> Tuple[List[Any], List[int]]:
    """各車両の総コストに基づく貪欲法で修復.
    
    各位置で最も良い値を選択。
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    base_cur : Sequence[Any]
        現在の基本解
    partial_x : Sequence[Any]
        部分解（削除位置はNone）
    removed : List[int]
        削除されたインデックス
    
    Returns
    -------
    Tuple[List[Any], List[int]]
        (修復された解, 変更されたインデックス) のタプル
    """
    W = state.additional_info.get("W")
    veh_keys = list(W.VEHICLES.keys())
    
    x = _copy_vec(base_cur)
    changed = []
    
    for j in removed:
        orig = x[j]
        veh = W.VEHICLES[veh_keys[j]]
        t = veh.departure_time_in_second
        o, d = veh.orig.name, veh.dest.name

        cost_min = float("inf")
        for v in state.domains[j]:
            x[j] = v
            route = W.defRoute(W.dict_od_to_routes[o, d][v])
            ext = estimate_congestion_externality_route(W, route, t)
            private_cost = route.actual_travel_time(t)
            cost = private_cost+ext
            
            if cost < cost_min:
                cost_min = cost
                best_val = v
        x[j] = best_val
        if x[j] != orig:
            changed.append(j)
    
    return x, changed


# 修復オペレータ一覧
_REPAIR_IMPLS: Dict[str, Callable[[ALNSState, Sequence[Any], Sequence[Any], List[int]], Tuple[List[Any], List[int]]]] = {
    "random": _repair_random,
    "min_cost": _repair_minimum_cost,
}

# 修復オペレータの初期重み（選択確率に比例）
_REPAIR_INITIAL_WEIGHTS: Dict[str, float] = {
    "random": 1.0,
    "min_cost": 1.0,
}

# -------------------------
# 初期化
# -------------------------
def init_alns(
    x: Sequence[Any],
    obj: Callable[[Sequence[Any]], float],
    *,
    sense: str = "min",
    domains: Optional[List[Sequence[Any]]] = None,

    # destroy/repair：指定しなければ辞書キーを自動採用
    destroy_set: Optional[List[str]] = None,
    repair_set: Optional[List[str]] = None,

    # 破壊規模
    k_min: int = None,
    k_max: int = 5,

    # 焼きなまし
    p0: float=0.2, #目標初期エラー
    trials: int=50, #初期試行回数
    T0: Optional[float] = None,
    Tend: float = 1e-3,
    total_iters: int = 2000,
    always_accept: False,

    # 適応（セグメント更新）
    segment_len: int = 50,
    adapt_eta: float = 0.2,
    reward_improve_best: float = 5.0,
    reward_improve: float = 2.0,
    reward_accept_worse: float = 0.5,

    seed: Optional[int] = 42,

    # 追加情報（任意）
    additional_info: Optional[Dict[str, Any]] = None,
) -> ALNSState:
    """ALNSアルゴリズムを初期化.
    
    Parameters
    ----------
    x : Sequence[Any]
        初期解
    obj : Callable[[Sequence[Any]], float]
        目的関数
    sense : str, default="min"
        最適化方向 ("min" または "max")
    domains : Optional[List[Sequence[Any]]], default=None
        各変数の定義域。Noneの場合はxから自動推定
    destroy_set : Optional[List[str]], default=None
        使用する破壊オペレータ名。Noneの場合は全て使用
    repair_set : Optional[List[str]], default=None
        使用する修復オペレータ名。Noneの場合は全て使用
    k_min : int, default=1
        破壊規模の最小値
    k_max : int, default=5
        破壊規模の最大値
    T0 : Optional[float], default=None
        初期温度。Noneの場合は自動推定
    Tend : float, default=1e-3
        終了温度
    total_iters : int, default=2000
        総イテレーション数
    segment_len : int, default=50
        適応更新のセグメント長
    adapt_eta : float, default=0.2
        適応更新の学習率
    reward_improve_best : float, default=5.0
        最良解改善時の報酬
    reward_improve : float, default=2.0
        解改善時の報酬
    reward_accept_worse : float, default=0.5
        悪化解受理時の報酬
    seed : Optional[int], default=42
        乱数シード
    additional_info : Optional[Dict[str, Any]], default=None
        追加情報（例: departure_timesなど）
    
    Returns
    -------
    ALNSState
        初期化されたALNS状態
    
    Raises
    ------
    ValueError
        senseが"min"または"max"でない場合
        domainsに空の位置がある場合
        destroy_setまたはrepair_setが空の場合
        未登録のオペレータが指定された場合
    """
    if sense not in ("min", "max"):
        raise ValueError("sense must be 'min' or 'max'")

    rng = random.Random(seed)

    if domains is None:
        domains = _infer_domains_from_x(x)
    n = len(x)
    if any(len(d) == 0 for d in domains):
        raise ValueError("domains is invalid")

    # 利用可能オペレータ一覧（辞書キー）
    available_destroy = list(_DESTROY_IMPLS.keys())
    available_repair  = list(_REPAIR_IMPLS.keys())

    if destroy_set is None:
        destroy_set = available_destroy[:]
    if repair_set is None:
        repair_set = available_repair[:]

    if not destroy_set:
        raise ValueError("destroy_set is empty")
    if not repair_set:
        raise ValueError("repair_set is empty")

    unknown_d = [m for m in destroy_set if m not in _DESTROY_IMPLS]
    unknown_r = [m for m in repair_set  if m not in _REPAIR_IMPLS]
    if unknown_d:
        raise ValueError(f"destroy_set is invalid")
    if unknown_r:
        raise ValueError(f"repair_set is invalid")

    if k_min is None:
        k_min = k_max
    k_min = max(1, min(k_min, n))
    k_max = max(k_min, min(k_max, n))
    k_cur = k_max

    x_cur = _copy_vec(x)
    additional_info = additional_info or {}

    def _eval(xx: Sequence[Any]) -> float:
        v = obj(xx)
        return -v if sense == "max" else v

    cur = _eval(x_cur)
    best = cur
    x_best = _copy_vec(x_cur)

    # T0 自動推定
    if T0 is None:
        # diffs = []
        # trials = min(20, 100)
        # for _ in range(trials):
        #     xx = _copy_vec(x_cur)
        #     k = rng.randint(k_min, min(k_max, n))
        #     idx = rng.sample(range(n), k)
        #     for j in idx:
        #         choices = [c for c in domains[j] if c != xx[j]]
        #         if choices:
        #             xx[j] = _rng_choice(rng, choices)
        #     diffs.append(max(0.0, _eval(xx) - cur))
        # scale = (sorted(diffs)[len(diffs)//2] if diffs else 1.0) or 1.0
        # T0 = max(1e-6, 3.0 * scale)
            
        deltas = []
        if total_iters > trials:
            total_iters -= trials
        else:
            trials = int(total_iters*0.5)
            total_iters -= trials
            warnings.warn("total_iters is too small for initial T0 estimation")
            
        for _ in range(trials):
            xx = _copy_vec(x_cur)
            k = rng.randint(k_min, min(k_max, n))
            idx = rng.sample(range(n), k)
            for j in idx:
                choices = [c for c in domains[j] if c != xx[j]]
                if choices:
                    xx[j] = _rng_choice(rng, choices)
            d = _eval(xx) - cur
            if d > 0:               # 悪化のみ収集
                deltas.append(d)
        if not deltas:
            deltas = [1.0]
        med = sorted(deltas)[len(deltas)//2]
        p0 = p0                 # 初期に悪化を割合p0前後で受理したい：かなりずれる．．．
        T0 = max(1e-6, med / max(1e-12, math.log(1.0/p0)))

    # 適応初期値（初期重みは定数から取得、定義されていなければ1.0）
    d_weights = {m: _DESTROY_INITIAL_WEIGHTS.get(m, 1.0) for m in destroy_set}
    d_scores  = {m: 0.0 for m in destroy_set}
    d_uses    = {m: 0   for m in destroy_set}
    d_best_improves = {m: 0 for m in destroy_set}

    r_weights = {m: _REPAIR_INITIAL_WEIGHTS.get(m, 1.0) for m in repair_set}
    r_scores  = {m: 0.0 for m in repair_set}
    r_uses    = {m: 0   for m in repair_set}
    r_best_improves = {m: 0 for m in repair_set}

    params = dict(
        sense=sense,
        destroy_set=destroy_set, repair_set=repair_set,
        available_destroy=available_destroy, available_repair=available_repair,
        k_min=k_min, k_max=k_max, k_cur=k_cur,
        T0=T0, Tend=Tend, total_iters=total_iters,
        segment_len=segment_len, adapt_eta=adapt_eta,
        reward_improve_best=reward_improve_best, reward_improve=reward_improve,
        reward_accept_worse=reward_accept_worse,
        seed=seed
    )

    return ALNSState(
        sense=sense, obj=obj, domains=domains,
        destroy_set=destroy_set, repair_set=repair_set,
        k_min=k_min, k_max=k_max,k_cur=k_cur,
        always_accept=always_accept,
        T0=T0, Tend=Tend, total_iters=total_iters,
        segment_len=segment_len, adapt_eta=adapt_eta,
        reward_improve_best=reward_improve_best, reward_improve=reward_improve,
        reward_accept_worse=reward_accept_worse,
        rng=rng, seed=seed, params=params,
        additional_info=additional_info,
        x_cur=x_cur, x_best=x_best, cur=cur, best=best,
        it=0, last_best_it=0,
        d_weights=d_weights, d_scores=d_scores, d_uses=d_uses, d_best_improves=d_best_improves,
        r_weights=r_weights, r_scores=r_scores, r_uses=r_uses, r_best_improves=r_best_improves,
        seg_countdown=segment_len,
        steps=[]
    )

# -------------------------
# 温度・ルーレット・重み更新
# -------------------------
def _temperature(state: ALNSState, it_next: Optional[int] = None) -> float:
    """焼きなまし温度を計算.
    
    指数的に減少する温度スケジュール。
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    it_next : Optional[int], default=None
        次のイテレーション番号。Noneの場合は現在のitを使用
    
    Returns
    -------
    float
        現在の温度
    """
    it = state.it if it_next is None else it_next
    it = max(0, min(it, max(1, state.total_iters)))
    return state.T0 * ((state.Tend / state.T0) ** (it / max(1, state.total_iters)))

def _roulette_pick(rng: random.Random, weights: Dict[str, float]) -> str:
    """ルーレット選択.
    
    重みに比例した確率でキーを選択。
    
    Parameters
    ----------
    rng : random.Random
        乱数生成器
    weights : Dict[str, float]
        各キーの重み
    
    Returns
    -------
    str
        選択されたキー
    """
    total = sum(max(1e-12, w) for w in weights.values())
    r = rng.random() * total
    s = 0.0
    last = None
    for k, w in weights.items():
        s += max(1e-12, w)
        last = k
        if r <= s:
            return k
    return last or next(iter(weights))

def _segment_update_weights(state: ALNSState):
    """セグメント終了時にオペレータ重みを更新.
    
    スコアと使用回数に基づいて適応的に重みを調整。
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    """
    if state.seg_countdown > 0:
        return
    for m in state.destroy_set:
        uses = max(1, state.d_uses[m])
        target = max(1e-6, state.d_scores[m] / uses)
        w_old = state.d_weights[m]
        state.d_weights[m] = (1 - state.adapt_eta) * w_old + state.adapt_eta * target
        state.d_scores[m] = 0.0
        state.d_uses[m] = 0
    for m in state.repair_set:
        uses = max(1, state.r_uses[m])
        target = max(1e-6, state.r_scores[m] / uses)
        w_old = state.r_weights[m]
        state.r_weights[m] = (1 - state.adapt_eta) * w_old + state.adapt_eta * target
        state.r_scores[m] = 0.0
        state.r_uses[m] = 0
    state.seg_countdown = state.segment_len

def _reward_step(state: ALNSState, d: str, r: str, *, improved_best: bool, accepted: bool, delta: float):
    """ステップ結果に基づいてオペレータに報酬を付与.
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    d : str
        使用された破壊オペレータ名
    r : str
        使用された修復オペレータ名
    improved_best : bool
        最良解が改善されたか
    accepted : bool
        ムーブが受理されたか
    delta : float
        目的関数値の差分
    """
    if improved_best:
        state.d_scores[d] += state.reward_improve_best
        state.r_scores[r] += state.reward_improve_best
    elif accepted and delta < 0:
        state.d_scores[d] += state.reward_improve
        state.r_scores[r] += state.reward_improve
    elif accepted and delta > 0:
        state.d_scores[d] += state.reward_accept_worse
        state.r_scores[r] += state.reward_accept_worse

# -------------------------
# 自動ステップ（ALNS）
# -------------------------
def auto_step(state: ALNSState) -> StepLog:
    """ALNSの1ステップを実行.
    
    破壊・修復オペレータを選択し、現在解を変更、
    焼きなましに基づいて受理判定を行う。
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    
    Returns
    -------
    StepLog
        実行されたステップのログ
    """
    # kを計算
    temp_cur = _temperature(state, state.it)
    state.k_cur = int((state.k_max-state.k_min)*temp_cur/state.T0 + state.k_min)
    if state.k_cur < state.k_min:
        state.k_cur = state.k_min
    if state.k_cur > state.k_max:
        state.k_cur = state.k_max

    # オペレータ選択
    d = _roulette_pick(state.rng, state.d_weights)
    r = _roulette_pick(state.rng, state.r_weights)
    state.d_uses[d] += 1
    state.r_uses[r] += 1

    # 破壊
    removed, partial = _DESTROY_IMPLS[d](state, state.x_cur)
    # 修復
    x_cand, changed_idx = _REPAIR_IMPLS[r](state, state.x_cur, partial, removed)

    # 評価・受理
    cand = _eval_internal(state, x_cand)
    delta = cand - state.cur
    T = _temperature(state, it_next=state.it + 1)

    accept = (((delta <= 0) or (random.random() < math.exp(-delta / max(T, 1e-12)))) or state.always_accept) and len(changed_idx) > 0
    if accept:
        state.x_cur, state.cur = x_cand, cand

    improved_best = False
    if state.cur < state.best:
        state.best = state.cur
        state.x_best = _copy_vec(state.x_cur)
        improved_best = True
        state.last_best_it = state.it + 1
        state.d_best_improves[d] += 1
        state.r_best_improves[r] += 1

    _reward_step(state, d, r, improved_best=improved_best, accepted=accept, delta=delta)

    state.it += 1
    state.seg_countdown -= 1
    _segment_update_weights(state)

    log = StepLog(
        iter=state.it, move=f"D:{d}+R:{r}", changed_idx=changed_idx, changed_idxs=len(changed_idx),
        delta_obj=delta, accepted=accept,
        obj_val=(state.cur if state.sense == "min" else -state.cur),
        best_obj=(state.best if state.sense == "min" else -state.best),
        temperature=T, source="auto"
    )
    state.steps.append(log)

    return log

def run_auto(state: ALNSState, n: int = 100, time_limit_s: Optional[float] = None) -> List[StepLog]:
    """複数ステップを自動実行.
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    n : int, default=100
        実行するステップ数
    time_limit_s : Optional[float], default=None
        時間制限（秒）。Noneの場合は制限なし
    
    Returns
    -------
    List[StepLog]
        実行されたステップのログリスト
    """
    start_it = state.it
    t0 = time.time()
    while state.it - start_it < n:
        if time_limit_s is not None and (time.time() - t0) >= time_limit_s:
            break
        auto_step(state)
    return state.steps[start_it:state.it]

# -------------------------
# 取得・評価・完了
# -------------------------
def get_current_x(state: ALNSState) -> List[Any]:
    """現在解を取得.
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    
    Returns
    -------
    List[Any]
        現在解のコピー
    """
    return _copy_vec(state.x_cur)

def get_best_x(state: ALNSState) -> List[Any]:
    """最良解を取得.
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    
    Returns
    -------
    List[Any]
        最良解のコピー
    """
    return _copy_vec(state.x_best)

# def evaluate_candidates(state: ALNSState, xs: Iterable[Sequence[Any]]) -> List[float]:
#     """複数の候補解を評価.
    
#     Parameters
#     ----------
#     state : ALNSState
#         ALNS状態
#     xs : Iterable[Sequence[Any]]
#         評価する候補解のイテラブル
    
#     Returns
#     -------
#     List[float]
#         各候補解の目的関数値（ユーザー向き）
#     """
#     vals = []
#     for x in xs:
#         v = state.obj(x)
#         vals.append(v if state.sense == "min" else -v)
#     return vals

def finalize(state: ALNSState, stop_reason: str = "") -> Tuple[Sequence[Any], RunLog]:
    """ALNS実行を終了し、結果を取得.
    
    Parameters
    ----------
    state : ALNSState
        ALNS状態
    stop_reason : str, default=""
        停止理由
    
    Returns
    -------
    Tuple[Sequence[Any], RunLog]
        (最良解, 実行ログ) のタプル
    """
    run = RunLog(
        steps=list(state.steps),
        best_x=_copy_vec(state.x_best),
        best_obj=(state.best if state.sense == "min" else -state.best),
        seed=state.seed,
        stop_reason=stop_reason,
        operator_stats={
            "destroy": {
                m: dict(weight=state.d_weights[m], uses=state.d_uses[m], best_improves=state.d_best_improves[m])
                for m in state.destroy_set
            },
            "repair": {
                m: dict(weight=state.r_weights[m], uses=state.r_uses[m], best_improves=state.r_best_improves[m])
                for m in state.repair_set
            },
        },
        params=dict(state.params),
    )
    return _copy_vec(state.x_best), run

# -------------------------
# デモ（__main__）
# -------------------------
# if __name__ == "__main__":
#     # 例：各位置 ∈ {0,1,2,3}、重み付き合計を target に合わせる（制約なし）
#     random.seed(42)
#     n_elem = 20
#     w = [random.uniform(0,10) for _ in range(n_elem)]
#     target = sum(w)*1.5
#     x0 = [0 for _ in range(n_elem)]
#     domains = [[0,1,2,3] for _ in x0]

#     def obj_example(x1):
#         return abs(sum(wj*xj for wj, xj in zip(w, x1)) - target)

#     state = init_alns(
#         x0, obj_example,
#         domains=domains,
#         k_min=1, k_max=4,
#         total_iters=2000,
#         segment_len=50,
#         seed=0,
#         additional_info={"departure_times": np.linspace(0,1,len(x0))} #無意味な例
#     )

#     # 1000 ステップだけ走らせて終了
#     run_auto(state, n=1000)
#     x_star, run = finalize(state, stop_reason="done")

#     try:
#         df = run.to_dataframe()
#         print(df.to_string())
#     except Exception:
#         pass

#     print("best obj:", run.best_obj)
#     print("best x  :", x_star)
#     print("op stats:")
#     pprint(run.operator_stats)
